"""
PULIRE

Validator Class

The code is licensed under the MIT license.
"""

from inspect import isfunction, signature
from typing import Callable, Union
from pandas import DataFrame, Series


class Validator:
    """
    Series Validator
    """

    func: Union[Callable, None] = None
    vectorized: bool = True
    skip_null: bool = False

    def __init__(self, func: Callable, vectorized: bool = True, skip_null=True):
        self.func = func
        self.vectorized = vectorized
        self.skip_null = skip_null

    def validate(
        self, series: Series, df: DataFrame, column: str
    ) -> Union[bool, Series]:
        """
        Run validator

        Returns a bool series:
        True -> Check passed
        False -> Check failed
        """
        arg_count = len((signature(self.func)).parameters)
        args = [series, df, column]
        if self.vectorized:
            return self.func(*args[0:arg_count])
        raise NotImplementedError("Pulire doesn't support non-vectorized checks, yet.")


def validate_df(validator: Validator, df: DataFrame, column: str) -> Series:
    """
    Validate a DataFrame's column using a validator
    """
    validator = validator() if isfunction(validator) else validator
    if validator.skip_null:
        result = Series(data=True, index=df.index, dtype=bool)
        result.update(
            validator.validate(
                df.loc[df[column].notnull()][column],
                df.loc[df[column].notnull()],
                column,
            )
        )
        return result.astype(bool)
    return validator.validate(df[column], df, column)


def validate_series(validator: Validator, series: Series) -> Series:
    """
    Validate a Series using a validator
    """
    validator = validator() if isfunction(validator) else validator
    if validator.skip_null:
        result = Series(data=True, index=series.index, dtype=bool)
        result.update(validator.validate(series.loc[series.notnull()], None, None))
        return result.astype(bool)
    return validator.validate(series, None, None)



"""
PULIRE

Schema Class

The code is licensed under the MIT license.
"""

from copy import copy
from typing import List
from pandas import DataFrame, Series, to_numeric

from pulire.column import Column
from pulire.validator import validate_df, validate_series
from pulire.formatter import format_df, format_series


class Schema:
    """
    Pulire Schema
    """

    is_series_schema = False
    columns: List[Column]

    def __init__(self, columns: List[Column] | Column):
        self.is_series_schema = isinstance(columns, Column)
        self.columns = [columns] if isinstance(columns, Column) else columns

    @property
    def dtypes(self) -> dict | str:
        """
        Dictionary of the column's data types
        """
        return self.columns[0].dtype if self.is_series_schema else {col.name: col.dtype for col in self.columns}

    def fit(self, data: DataFrame | Series, fill=None) -> DataFrame:
        """
        Remove invalid data from a DataFrame, enforce data types and apply formatters
        """
        temp = copy(data)

        # Rename series
        if self.is_series_schema:
            temp.rename(self.columns[0].name, inplace=True)

        # Enforce data types
        if self.is_series_schema:
            if self.columns[0].dtype == "Int64":
                temp = to_numeric(temp).round(0)
            temp = temp.astype(self.columns[0].dtype, errors="ignore")
        else:
            for col, dtype in self.dtypes.items():
                if col in temp and dtype == "Int64":
                    temp[col] = to_numeric(temp[col]).round(0)
            temp = temp.astype(self.dtypes, errors="ignore")

        # Remove invalid data
        if self.is_series_schema:
            for validator in self.columns[0].validators:
                test = validate_series(validator, temp)
                temp.loc[~test] = fill
        else:
            for col in self.columns:
                if col.name in temp.columns:
                    for validator in col.validators:
                        test = validate_df(validator, temp, col.name)
                        temp.loc[~test, col.name] = fill

        # Apply formatters
        if self.is_series_schema:
            for formatter in self.columns[0].formatters:
                temp = format_series(formatter, temp)
        else:
            for col in self.columns:
                if col.name in temp.columns:
                    for formatter in col.formatters:
                        temp[col.name] = format_df(formatter, temp, col.name)

        return temp

    def validate(self, df: DataFrame) -> None:
        """
        Raise error when checks are failing
        """
        for col in self.columns:
            if col.name in df.columns:
                for validator in col.validators:
                    test = validate_df(validator, df, col.name)
                    if not test.all():
                        raise ValueError(f'Column "{col.name}" contains invalid data')

    def is_valid(self, df: DataFrame) -> bool:
        """
        Check if a DataFrame is valid
        """
        for col in self.columns:
            if col.name in df.columns:
                for validator in col.validators:
                    test = validate_df(validator, df, col.name)
                    if not test.all():
                        return False
        return True



"""
PULIRE

The code is licensed under the MIT license.
"""

__appname__ = "pulire"
__version__ = "0.0.4"

from . import validators, formatters
from .validator import Validator
from .schema import Schema
from .column import Column

__all__ = ["validators", "formatters", "Validator", "Schema", "Column"]



"""
Formatter Class

The code is licensed under the MIT license.
"""

from inspect import isfunction, signature
from typing import Callable, Union
from pandas import DataFrame, Series


class Formatter:
    """
    Series Formatter
    """

    func: Union[Callable, None] = None

    def __init__(self, func: Callable, vectorized: bool = True, skip_null=True):
        self.func = func
        self.vectorized = vectorized
        self.skip_null = skip_null

    def format(self, series: Series, df: DataFrame, column: str) -> Series:
        """
        Format all values in a series
        """
        arg_count = len((signature(self.func)).parameters)
        args = [series, df, column]
        if self.vectorized:
            return self.func(*args[0:arg_count])
        raise NotImplementedError("Pulire doesn't support non-vectorized checks, yet.")


def format_df(formatter: Formatter, df: DataFrame, column: str) -> Series:
    """
    Format a DataFrame's column using a formatter
    """
    formatter = formatter() if isfunction(formatter) else formatter
    return formatter.format(df[column], df, column)


def format_series(formatter: Formatter, series: Series) -> Series:
    """
    Format a DataFrame's column using a formatter
    """
    formatter = formatter() if isfunction(formatter) else formatter
    return formatter.format(series, None, None)
